@@ -25,7 +25,7 @@ namespace Botan {
 *  - x86 features using CPUID. x86 is also the only processor with
 *    accurate cache line detection currently.
 *
-*  - PowerPC AltiVec detection on Linux, NetBSD, OpenBSD, and Darwin
+*  - PowerPC AltiVec detection on Linux, NetBSD, OpenBSD, and macOS
 *
 *  - ARM NEON and crypto extensions detection. On Linux and Android
 *    systems which support getauxval, that is used to access CPU
@@ -65,21 +65,17 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
       */
       static size_t cache_line_size()
          {
-         if(g_processor_features == 0)
-            {
-            initialize();
-            }
-         return g_cache_line_size;
+         return state().cache_line_size();
          }
 
       static bool is_little_endian()
          {
-         return endian_status() == ENDIAN_LITTLE;
+         return state().endian_status() == Endian_Status::Little;
          }
 
       static bool is_big_endian()
          {
-         return endian_status() == ENDIAN_BIG;
+         return state().endian_status() == Endian_Status::Big;
          }
 
       enum CPUID_bits : uint64_t {
@@ -98,6 +94,7 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
          CPUID_RDTSC_BIT   = (1ULL << 10),
          CPUID_BMI2_BIT    = (1ULL << 11),
          CPUID_ADX_BIT     = (1ULL << 12),
+         CPUID_BMI1_BIT    = (1ULL << 13),
 
          // Crypto-specific ISAs
          CPUID_AESNI_BIT   = (1ULL << 16),
@@ -108,15 +105,21 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
 #endif
 
 #if defined(BOTAN_TARGET_CPU_IS_PPC_FAMILY)
-         CPUID_ALTIVEC_BIT = (1ULL << 0),
+         CPUID_ALTIVEC_BIT    = (1ULL << 0),
+         CPUID_PPC_CRYPTO_BIT = (1ULL << 1),
 #endif
 
 #if defined(BOTAN_TARGET_CPU_IS_ARM_FAMILY)
-         CPUID_ARM_NEON_BIT  = (1ULL << 0),
-         CPUID_ARM_AES_BIT   = (1ULL << 16),
-         CPUID_ARM_PMULL_BIT = (1ULL << 17),
-         CPUID_ARM_SHA1_BIT  = (1ULL << 18),
-         CPUID_ARM_SHA2_BIT  = (1ULL << 19),
+         CPUID_ARM_NEON_BIT      = (1ULL << 0),
+         CPUID_ARM_SVE_BIT       = (1ULL << 1),
+         CPUID_ARM_AES_BIT       = (1ULL << 16),
+         CPUID_ARM_PMULL_BIT     = (1ULL << 17),
+         CPUID_ARM_SHA1_BIT      = (1ULL << 18),
+         CPUID_ARM_SHA2_BIT      = (1ULL << 19),
+         CPUID_ARM_SHA3_BIT      = (1ULL << 20),
+         CPUID_ARM_SHA2_512_BIT  = (1ULL << 21),
+         CPUID_ARM_SM3_BIT       = (1ULL << 22),
+         CPUID_ARM_SM4_BIT       = (1ULL << 23),
 #endif
 
          CPUID_INITIALIZED_BIT = (1ULL << 63)
@@ -128,6 +131,13 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
       */
       static bool has_altivec()
          { return has_cpuid_bit(CPUID_ALTIVEC_BIT); }
+
+      /**
+      * Check if the processor supports POWER8 crypto extensions
+      */
+      static bool has_ppc_crypto()
+         { return has_cpuid_bit(CPUID_PPC_CRYPTO_BIT); }
+
 #endif
 
 #if defined(BOTAN_TARGET_CPU_IS_ARM_FAMILY)
@@ -137,6 +147,12 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
       static bool has_neon()
          { return has_cpuid_bit(CPUID_ARM_NEON_BIT); }
 
+      /**
+      * Check if the processor supports ARMv8 SVE
+      */
+      static bool has_arm_sve()
+         { return has_cpuid_bit(CPUID_ARM_SVE_BIT); }
+
       /**
       * Check if the processor supports ARMv8 SHA1
       */
@@ -160,6 +176,31 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
       */
       static bool has_arm_pmull()
          { return has_cpuid_bit(CPUID_ARM_PMULL_BIT); }
+
+      /**
+      * Check if the processor supports ARMv8 SHA-512
+      */
+      static bool has_arm_sha2_512()
+         { return has_cpuid_bit(CPUID_ARM_SHA2_512_BIT); }
+
+      /**
+      * Check if the processor supports ARMv8 SHA-3
+      */
+      static bool has_arm_sha3()
+         { return has_cpuid_bit(CPUID_ARM_SHA3_BIT); }
+
+      /**
+      * Check if the processor supports ARMv8 SM3
+      */
+      static bool has_arm_sm3()
+         { return has_cpuid_bit(CPUID_ARM_SM3_BIT); }
+
+      /**
+      * Check if the processor supports ARMv8 SM4
+      */
+      static bool has_arm_sm4()
+         { return has_cpuid_bit(CPUID_ARM_SM4_BIT); }
+
 #endif
 
 #if defined(BOTAN_TARGET_CPU_IS_X86_FAMILY)
@@ -206,6 +247,12 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
       static bool has_avx512f()
          { return has_cpuid_bit(CPUID_AVX512F_BIT); }
 
+      /**
+      * Check if the processor supports BMI1
+      */
+      static bool has_bmi1()
+         { return has_cpuid_bit(CPUID_BMI1_BIT); }
+
       /**
       * Check if the processor supports BMI2
       */
@@ -258,8 +305,7 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
       */
       static void clear_cpuid_bit(CPUID_bits bit)
          {
-         const uint64_t mask = ~(static_cast<uint64_t>(bit));
-         g_processor_features &= mask;
+         state().clear_cpuid_bit(static_cast<uint64_t>(bit));
          }
 
       /*
@@ -268,41 +314,56 @@ class BOTAN_PUBLIC_API(2,1) CPUID final
       */
       static bool has_cpuid_bit(CPUID_bits elem)
          {
-         if(g_processor_features == 0)
-            initialize();
-         return ((g_processor_features & static_cast<uint64_t>(elem)) != 0);
+         const uint64_t elem64 = static_cast<uint64_t>(elem);
+         return state().has_bit(elem64);
          }
 
       static std::vector<CPUID::CPUID_bits> bit_from_string(const std::string& tok);
    private:
-      enum Endian_status : uint32_t {
-         ENDIAN_UNKNOWN = 0x00000000,
-         ENDIAN_BIG     = 0x01234567,
-         ENDIAN_LITTLE  = 0x67452301,
+      enum class Endian_Status : uint32_t {
+         Unknown = 0x00000000,
+         Big     = 0x01234567,
+         Little  = 0x67452301,
       };
 
+      struct CPUID_Data
+         {
+         public:
+            CPUID_Data();
+
+            CPUID_Data(const CPUID_Data& other) = default;
+            CPUID_Data& operator=(const CPUID_Data& other) = default;
+
+            void clear_cpuid_bit(uint64_t bit)
+               {
+               m_processor_features &= ~bit;
+               }
+
+            bool has_bit(uint64_t bit) const
+               {
+               return (m_processor_features & bit) == bit;
+               }
+
+            uint64_t processor_features() const { return m_processor_features; }
+            Endian_Status endian_status() const { return m_endian_status; }
+            size_t cache_line_size() const { return m_cache_line_size; }
+
+         private:
+            static Endian_Status runtime_check_endian();
+
 #if defined(BOTAN_TARGET_CPU_IS_PPC_FAMILY) || \
     defined(BOTAN_TARGET_CPU_IS_ARM_FAMILY) || \
     defined(BOTAN_TARGET_CPU_IS_X86_FAMILY)
 
-      static uint64_t detect_cpu_features(size_t* cache_line_size);
+            static uint64_t detect_cpu_features(size_t* cache_line_size);
 
 #endif
+            uint64_t m_processor_features;
+            size_t m_cache_line_size;
+            Endian_Status m_endian_status;
+         };
 
-      static Endian_status runtime_check_endian();
-
-      static Endian_status endian_status()
-         {
-         if(g_endian_status == ENDIAN_UNKNOWN)
-            {
-            g_endian_status = runtime_check_endian();
-            }
-         return g_endian_status;
-         }
-
-      static uint64_t g_processor_features;
-      static size_t g_cache_line_size;
-      static Endian_status g_endian_status;
+      static CPUID_Data& state();
    };
 
 }
